/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz>
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "mozCellML.h"
#include "mozCellMLVariable.h"
#include "mozCellMLIntegrator.h"
#include "mozCellMLChannel.h"
#include <nsIURL.h>
#include <necko/nsNetCID.h>
#include <nsIServiceManager.h>
#include <nsCOMArray.h>
#include <nsArray.h>
#include <nsArrayEnumerator.h>
#include <nsIDOMDocument.h>
#include <necko/nsIServerSocket.h>
#include <nsDirectoryService.h>
#include <nsPrintfCString.h>
#include <nsContentCID.h>
#include <nsISyncLoadDOMService.h>
#include <necko/nsNetUtil.h>
#include <nsIContent.h>

NS_IMPL_ISUPPORTS1(mozCellML, mozICellML)

mozCellML::mozCellML()
  : localValidationError(PR_FALSE),
    validationErrorMessage(NS_LITERAL_STRING(""))
{
  NS_NewArray(getter_AddRefs(variableList));
  NS_NewArray(getter_AddRefs(allComponents));
  allVariables.Init();
}

mozCellML::~mozCellML()
{
}

NS_IMETHODIMP
mozCellML::Initialise()
{
  PRUint32 i, l;
  if (doc == nsnull)
  {
    printf("Initialise called but document is null.\n");
    return NS_ERROR_FAILURE;
  }

  nsCOMPtr<nsIDOMElement> de;
  doc->GetDocumentElement(getter_AddRefs(de));
  if (de == nsnull)
  {
    printf("DocumentElement for document is null.\n");
    return NS_ERROR_FAILURE;
  }

  FindVariableElements(de);

  // We need to search connections to link variables across components...
  nsCOMPtr<nsIDOMNodeList> dnl;
  nsCOMPtr<nsIArray> connections;
  de->GetChildNodes(getter_AddRefs(dnl));
  nsCOMPtr<nsISimpleEnumerator> connectionsE;
  FilterNodeListByNameAndTypeNS2(
                                 dnl,
                                 getter_AddRefs(connections),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("connection")
                                );
  connections->GetLength(&l);
  connections->Enumerate(getter_AddRefs(connectionsE));

  nsInterfaceHashtable<nsStringHashKey, mozCellMLVariablePromise>
    PromiseTable;
  PromiseTable.Init();
  nsCOMArray<mozCellMLVariablePromise> PromiseList;
  
  for (
       i = 0;
       i < l;
       i++
      )
  {
    nsCOMPtr<nsISupports> connSupports;
    connectionsE->GetNext(getter_AddRefs(connSupports));
    nsCOMPtr<nsIDOMElement> conn(do_QueryInterface(connSupports));
    allConnections.AppendObject(conn);
  }

  l = allConnections.Count();
  for (
       i = 0;
       i < l;
       i++
      )
  {
    nsIDOMElement* conn = allConnections[i];
    nsCOMPtr<nsIArray> mapcomps;
    conn->GetChildNodes(getter_AddRefs(dnl));
    FilterNodeListByNameAndTypeNS2(
                                   dnl,
                                   getter_AddRefs(mapcomps),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("map_components")
                                  );
    PRUint32 nmc, j;
    mapcomps->GetLength(&nmc);
    if (nmc == 0)
      continue;
    nsCOMPtr<nsIDOMElement> mapel;
    mapcomps->QueryElementAt(0, NS_GET_IID(nsIDOMElement),
                             getter_AddRefs(mapel));
    nsAutoString comp1, comp2;
    mapel->GetAttribute(NS_LITERAL_STRING("component_1"), comp1);
    mapel->GetAttribute(NS_LITERAL_STRING("component_2"), comp2);

    FilterNodeListByNameAndTypeNS2(
                                   dnl,
                                   getter_AddRefs(mapcomps),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("map_variables")
                                  );
    mapcomps->GetLength(&nmc);
    nsCOMPtr<nsISimpleEnumerator> varenum;
    mapcomps->Enumerate(getter_AddRefs(varenum));

    for (
         j = 0;
         j < nmc;
         j++
        )
    {
      nsCOMPtr<nsISupports> varSupports;
      varenum->GetNext(getter_AddRefs(varSupports));
      nsCOMPtr<nsIDOMElement> varel(do_QueryInterface(varSupports));
      nsAutoString var1, var2;
      varel->GetAttribute(NS_LITERAL_STRING("variable_1"), var1);
      varel->GetAttribute(NS_LITERAL_STRING("variable_2"), var2);
      nsAutoString map1(comp1 + NS_LITERAL_STRING("::") + var1),
                   map2(comp2 + NS_LITERAL_STRING("::") + var2);
      nsCOMPtr<mozCellMLVariable> var;
      allVariables.Get(map1, getter_AddRefs(var));
      if (var == nsnull)
      {
        allVariables.Get(map2, getter_AddRefs(var));
        if (var != NULL)
          allVariables.Put(map1, var);
        else
        {
          // See if we have a promise...
          nsCOMPtr<mozCellMLVariablePromise> promise1, promise2;
          PromiseTable.Get(map1, getter_AddRefs(promise1));
          PromiseTable.Get(map2, getter_AddRefs(promise2));
          if (promise1 != nsnull && promise2 != nsnull)
            promise1->MergeIn(promise2);
          else if (promise1 == nsnull && promise2 == nsnull)
          {
            promise1 = new mozCellMLVariablePromise(map1, map2);
            PromiseTable.Put(map1, promise1);
            PromiseTable.Put(map2, promise1);
            PromiseList.AppendObject(promise1);
          }
          else if (promise1 != nsnull)
          {
            promise1->AddParty(map2);
            PromiseTable.Put(map2, promise1);
          }
          else if (promise2 != nsnull)
          {
            promise2->AddParty(map1);
            PromiseTable.Put(map1, promise2);
          }
        }
      }
      else
        allVariables.Put(map2, var);
    }
  }
  nsresult rv;
  for (i = 0, l = PromiseList.Count(); i < l; i++)
  {
    rv = PromiseList.ObjectAt(i)->Resolve(this, allVariables);
    NS_ENSURE_SUCCESS(rv, rv);
  }

  // We now go through all variables and add the reference...
  nsCOMPtr<nsISimpleEnumerator> allEnum;
  variableList->Enumerate(getter_AddRefs(allEnum));
  variableList->GetLength(&l);
  nsCOMPtr<nsIArray> allComponentsA(do_QueryInterface(allComponents));
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> vSupports;
    allEnum->GetNext(getter_AddRefs(vSupports));
    nsCOMPtr<mozCellMLVariable> v(do_QueryInterface(vSupports));
    v->Initialise(allVariables, allComponentsA);
  }

  nsCOMPtr<nsISimpleEnumerator> varEnum;
  nsCOMPtr<mozCellMLVariable> var, bound;

  variableList->Enumerate(getter_AddRefs(varEnum));
  PRBool hme;
  while (varEnum->HasMoreElements(&hme), hme)
  {
    nsCOMPtr<nsISupports> varSupports;
    varEnum->GetNext(getter_AddRefs(varSupports));
    var = do_QueryInterface(varSupports);
    PRUint32 type;
    var->GetType(&type);
    if (type == mozCellMLIVariable::BOUND)
    {
      if (bound != nsnull)
        
      bound = var;
    }
    else if (type == mozCellMLIVariable::DIFFERENTIAL)
      differentials.AppendObject(var);
    else if (type == mozCellMLIVariable::SIMPLE)
      simples.AppendObject(var);
  }

  // Compute the dependencies now...
  for (
       i = 0, l = simples.Count();
       i < l;
       i++
      )
    simples[i]->PrecomputeDependencies(this, allVariables);
  
  // We are now ready to make a dependency resolution...
  for (
       i = 0, l = simples.Count();
       i < l;
       i++
      )
  {
    PRBool wasSeen;
    simples[i]->GetSeen(&wasSeen);
    if (wasSeen)
      continue;
    nsresult ret = simples[i]->OrderDependencies(this, order);
    NS_ENSURE_SUCCESS(ret, ret);
  }

  // And we name the variables...
  PRUint32 lt;
  for (
       i = 0, l = order.Count();
       i < l;
       i++
      )
  {
    nsPrintfCString name("x[%u]", i);
    order[i]->SetVariableName(name);
  }
  for (lt = l + differentials.Count();
       i < lt;
       i++)
  {
    nsPrintfCString name("x[%u]", i);
    differentials[i - l]->SetVariableName(name);
  }

  return NS_OK;
}

nsresult
mozCellML::ReadDocument(const nsAString& docURL, nsACString& errorMessage)
{
  static NS_DEFINE_CID(kXMLDocumentCID, NS_XMLDOCUMENT_CID);
  doc = do_CreateInstance(kXMLDocumentCID);
  if (doc == NULL)
  {
    errorMessage = "Can't make a document";
    return NS_ERROR_FAILURE;
  }

  nsCOMPtr<nsIURI> uri;
  nsCOMPtr<nsIChannel> chan;
  nsresult rv = NS_NewURI(getter_AddRefs(uri), docURL);
  NS_ENSURE_SUCCESS(rv, rv);
  rv = NS_NewChannel(getter_AddRefs(chan), uri);
  nsCOMPtr<nsISyncLoadDOMService>
    slds(do_GetService("@mozilla.org/content/syncload-dom-service;1"));
  rv = slds->LoadDocument(chan, nsnull, getter_AddRefs(doc));
  if (rv != NS_OK)
  {
    printf("Loading document failed: %08X\n", rv);
    return rv;
  }

  if (rv != NS_OK)
  {
    switch (rv)
    {
    case NS_ERROR_FAILURE:
      errorMessage = NS_LITERAL_CSTRING("Could not load file(generic failure)");
      break;
    default:
      errorMessage = NS_LITERAL_CSTRING("Could not load file.");
    }
    return NS_OK;
  }

#ifdef USE_CELLML_API
  model.fromNode(doc);
#endif

  rv = Initialise();
  NS_ENSURE_SUCCESS(rv, rv);

  errorMessage = NS_LITERAL_CSTRING("");

  return NS_OK;
}

nsresult
mozCellML::SetDom(nsIDOMDocument* dom)
{
  doc = dom;
  if (doc == NULL)
    return NS_ERROR_FAILURE;
#ifdef USE_CELLML_API
  model.fromNode(doc);
#endif
  return NS_OK;
}

nsresult
mozCellML::GetDom(nsIDOMDocument** dom)
{
  CallQueryInterface(doc, dom);
  return NS_OK;
}

nsresult
mozCellML::FetchImports(nsIMutableArray** allImports)
{
  nsCOMPtr<nsIDOMElement> de;
  doc->GetDocumentElement(getter_AddRefs(de));
  if (de == nsnull)
  {
    printf("DocumentElement for document is null.\n");
    return NS_ERROR_FAILURE;
  }

  nsCOMPtr<nsIDOMNodeList> dnl;
  de->GetChildNodes(getter_AddRefs(dnl));
  nsCOMPtr<nsIArray> imports;
  nsCOMPtr<nsISimpleEnumerator> importsE;
  // Imports are only CellML 1.1...
  FilterNodeListByNameAndTypeNS(
                                dnl,
                                getter_AddRefs(imports),
                                nsIDOMNode::ELEMENT_NODE,
                                CELLML_NS_1_1,
                                NS_LITERAL_STRING("import")
                               );
  imports->Enumerate(getter_AddRefs(importsE));
  PRUint32 i, l;

  NS_NewArray(allImports);

  for (i = 0, imports->GetLength(&l); i < l; i++)
  {
    nsCOMPtr<nsISupports> importSupports;
    importsE->GetNext(getter_AddRefs(importSupports));
    nsCOMPtr<nsIDOMElement> import(do_QueryInterface(importSupports));
    nsCOMPtr<mozCellMLImportRequest> mcir = new mozCellMLImportRequest(import);
    (*allImports)->AppendElement(mcir, PR_FALSE);
  }

  return NS_OK;
}

PRUint32
mozCellML::SearchDeeplyForEncapsulations(
                                         nsIDOMElement* searchIn,
                                         nsInterfaceHashtable<nsStringHashKey,
                                                              nsIDOMElement>
                                           &encapsulations
                                        )
{
  nsCOMPtr<nsIDOMNodeList> cl;
  searchIn->GetChildNodes(getter_AddRefs(cl));
  nsCOMPtr<nsIArray> allComponents;
  FilterNodeListByNameAndTypeNS2(
                                 cl,
                                 getter_AddRefs(allComponents),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("component_ref")
                                );
  PRUint32 i, l;
  allComponents->GetLength(&l);
  nsCOMPtr<nsISimpleEnumerator> allComponentsE;
  allComponents->Enumerate(getter_AddRefs(allComponentsE));
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> componentSupports;
    allComponentsE->GetNext(getter_AddRefs(componentSupports));
    nsCOMPtr<nsIDOMElement> componentEl(do_QueryInterface(componentSupports));
    nsAutoString compName;
    componentEl->GetAttribute(NS_LITERAL_STRING("component"), compName);
    if (SearchDeeplyForEncapsulations(componentEl, encapsulations) != 0)
      encapsulations.Put(compName, componentEl);
  }
  return l;
}

nsresult
mozCellML::ExpandEncapsulatedList(
                                  nsCOMArray<nsIDOMElement>& ComponentStack,
                                  nsInterfaceHashtable<nsStringHashKey,
                                                       nsIDOMElement>& AllComponents,
                                  nsInterfaceHashtable<nsStringHashKey,
                                                       nsIDOMElement>& imported,
                                  nsIDOMElement* importedRoot
                                 )
{
  // We have to do a deep search for component_refs and add them to
  // a hashtable by component name...
  nsInterfaceHashtable<nsStringHashKey,nsIDOMElement> encapsulations;
  encapsulations.Init();
  nsCOMPtr<nsIDOMNodeList> cl;
  importedRoot->GetChildNodes(getter_AddRefs(cl));
  nsCOMPtr<nsIArray> allGroups;
  FilterNodeListByNameAndTypeNS2(
                                 cl,
                                 getter_AddRefs(allGroups),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("group")
                                );
  nsCOMPtr<nsISimpleEnumerator> allGroupsE;
  allGroups->Enumerate(getter_AddRefs(allGroupsE));
  PRUint32 i, l;
  allGroups->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> groupSupports;
    allGroupsE->GetNext(getter_AddRefs(groupSupports));
    nsCOMPtr<nsIDOMElement> groupEl(do_QueryInterface(groupSupports));
    PRBool encapsulation = PR_FALSE;
    importedRoot->GetChildNodes(getter_AddRefs(cl));
    nsCOMPtr<nsIArray> allRelationships;
    FilterNodeListByNameAndTypeNS2(
                                   cl,
                                   getter_AddRefs(allRelationships),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("relationship_ref")
                                  );
    nsCOMPtr<nsISimpleEnumerator> allRelationshipsE;
    allRelationships->Enumerate(getter_AddRefs(allRelationshipsE));
    PRUint32 m, j;
    allRelationships->GetLength(&m);
    for (j = 0; j < m; j++)
    {
      nsCOMPtr<nsISupports> relSupports;
      allRelationshipsE->GetNext(getter_AddRefs(relSupports));
      nsCOMPtr<nsIDOMElement> relEl(do_QueryInterface(relSupports));
      nsAutoString relationship;
      relEl->GetAttribute(NS_LITERAL_STRING("relationship"), relationship);
      if (relationship.Equals(NS_LITERAL_STRING("encapsulation")))
      {
        encapsulation = PR_TRUE;
        break;
      }
    }
    if (!encapsulation)
      continue;
    SearchDeeplyForEncapsulations(groupEl, encapsulations);
  }

  nsCOMPtr<nsIAtom> seenAtom = NS_NewAtom("seen_for_import");

  // Keep processing things on the stack...
  while ((l = ComponentStack.Count()) > 0)
  {
    nsIDOMElement* componentEl = ComponentStack[l - 1];
    ComponentStack.RemoveObjectAt(l - 1);
    // Get the name...
    nsAutoString compName;
    componentEl->GetAttribute(NS_LITERAL_STRING("name"), compName);

    // Check the hashtable...
    nsCOMPtr<nsIDOMElement> compRefEl;
    encapsulations.Get(compName, getter_AddRefs(compRefEl));
    if (compRefEl == nsnull)
      continue;
    // We have other components encapsulated in this one...
    compRefEl->GetChildNodes(getter_AddRefs(cl));
    nsCOMPtr<nsIArray> allSubComponents;
    FilterNodeListByNameAndTypeNS2(
                                   cl,
                                   getter_AddRefs(allSubComponents),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("relationship_ref")
                                  );
    nsCOMPtr<nsISimpleEnumerator> allSubsE;
    allSubComponents->Enumerate(getter_AddRefs(allSubsE));
    allSubComponents->GetLength(&l);
    PRUint32 i;
    for (i = 0; i < l; i++)
    {
      nsCOMPtr<nsISupports> compSupports;
      allSubsE->GetNext(getter_AddRefs(compSupports));
      nsCOMPtr<nsIDOMElement> compRefEl(do_QueryInterface(compSupports));
      compRefEl->GetAttribute(NS_LITERAL_STRING("component"), compName);
      // Find the real component...
      nsCOMPtr<nsIDOMElement> component;
      AllComponents.Get(compName, getter_AddRefs(component));
      // XXX we need to try a recursive import...
      if (component == nsnull)
      {
        return ValidationError(
                NS_LITERAL_STRING("Imported model does not have component ") +
                compName +
                NS_LITERAL_STRING(" but it is mentioned in encapsulation group.")
                              );
      }
      imported.Put(compName, component);

      // Set the component as seen...
#if 0 // The API isn't stable. For now, trust authors not to make loops...
      nsCOMPtr<nsIContent> componentNode(do_QueryInterface(component));
      nsresult rv;
      if (componentNode->GetProperty(seenAtom) != nsnull)
        continue;
      componentNode->SetProperty(seenAtom, this, nsnull);
#endif
      // Add it to the queue...
      ComponentStack.AppendObject(component);
      // Install it...
      InstallComponent(component, compName);
    }
  }
  return NS_OK;
}

nsresult
mozCellML::HashAvailableComponents(nsInterfaceHashtable<nsStringHashKey,
                                                        nsIDOMElement>& addTo,
                                   nsIDOMElement* importedRoot)
{
  nsCOMPtr<nsIDOMNodeList> cl;
  importedRoot->GetChildNodes(getter_AddRefs(cl));
  nsCOMPtr<nsIArray> gotComponents;
  FilterNodeListByNameAndTypeNS2(
                                 cl,
                                 getter_AddRefs(gotComponents),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("component")
                                );
  nsCOMPtr<nsISimpleEnumerator> gotComponentsE;
  gotComponents->Enumerate(getter_AddRefs(gotComponentsE));
  PRUint32 i, l;
  gotComponents->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> componentSupports;
    gotComponentsE->GetNext(getter_AddRefs(componentSupports));
    nsCOMPtr<nsIDOMElement> gotComponent(do_QueryInterface(componentSupports));
    nsAutoString compName;
    gotComponent->GetAttribute(NS_LITERAL_STRING("name"), compName);
    addTo.Put(compName, gotComponent);
  }
  return NS_OK;
}

nsresult
mozCellML::AddDirectImports(
                            nsInterfaceHashtable<nsStringHashKey,
                                                 nsIDOMElement>& available,
                            nsInterfaceHashtable<nsStringHashKey,
                                                 nsIDOMElement>& imported,
                            nsIDOMElement* importNode,
                            nsCOMArray<nsIDOMElement>& importStack
                           )
{
  nsCOMPtr<nsIDOMNodeList> cl;
  importNode->GetChildNodes(getter_AddRefs(cl));
  nsCOMPtr<nsIArray> wantComponents;

  // Get all the CellML components listed inside the import...
  FilterNodeListByNameAndTypeNS(
                                cl,
                                getter_AddRefs(wantComponents),
                                nsIDOMNode::ELEMENT_NODE,
                                CELLML_NS_1_1,
                                NS_LITERAL_STRING("component")
                               );

  PRUint32 i, l;
  nsCOMPtr<nsISimpleEnumerator> wantComponentsE;
  wantComponents->Enumerate(getter_AddRefs(wantComponentsE));
  for (i = 0, wantComponents->GetLength(&l); i < l; i++)
  {
    nsCOMPtr<nsISupports> componentSupports;
    wantComponentsE->GetNext(getter_AddRefs(componentSupports));
    nsCOMPtr<nsIDOMElement> wantComponent(do_QueryInterface(componentSupports));
    nsAutoString wantName, wantActualName;
    wantComponent->GetAttribute(NS_LITERAL_STRING("name"), wantName);
    wantComponent->GetAttribute(NS_LITERAL_STRING("component_ref"),
                                wantActualName);

    nsCOMPtr<nsIDOMElement> gotComponent;
    available.Get(wantActualName, getter_AddRefs(gotComponent));
    imported.Put(wantActualName, gotComponent);
    if (gotComponent != nsnull)
      InstallComponent(gotComponent, wantName);
    else
      return ValidationError(
                 NS_LITERAL_STRING("Imported model does not have component ") +
                 wantActualName +
                 NS_LITERAL_STRING(" needed to add component ") + wantName
                            );
  }
  return NS_OK;
}

nsresult
mozCellML::AddRelevantConnections(nsIDOMElement* importedRoot,
                                  nsInterfaceHashtable<nsStringHashKey,
                                                       nsIDOMElement>
                                  &imported)
{
  nsCOMPtr<nsIDOMNodeList> cl;
  importedRoot->GetChildNodes(getter_AddRefs(cl));
  nsCOMPtr<nsIArray> connections;
  // Get all the CellML components listed inside the import...
  FilterNodeListByNameAndTypeNS2(
                                 cl,
                                 getter_AddRefs(connections),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("connection")
                                );

  PRUint32 i, l;
  nsCOMPtr<nsISimpleEnumerator> connectionsE;
  connections->GetLength(&l);
  connections->Enumerate(getter_AddRefs(connectionsE));

  for (
       i = 0;
       i < l;
       i++
      )
  {
    nsCOMPtr<nsISupports> connSupports;
    connectionsE->GetNext(getter_AddRefs(connSupports));
    nsCOMPtr<nsIDOMElement> conn(do_QueryInterface(connSupports));

    nsCOMPtr<nsIArray> mapcomps;
    conn->GetChildNodes(getter_AddRefs(cl));
    FilterNodeListByNameAndTypeNS2(
                                   cl,
                                   getter_AddRefs(mapcomps),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("map_components")
                                  );
    PRUint32 nmc;
    mapcomps->GetLength(&nmc);
    if (nmc == 0)
      continue;
    nsCOMPtr<nsIDOMElement> mapel;
    mapcomps->QueryElementAt(0, NS_GET_IID(nsIDOMElement),
                             getter_AddRefs(mapel));
    nsAutoString comp1, comp2;
    mapel->GetAttribute(NS_LITERAL_STRING("component_1"), comp1);
    mapel->GetAttribute(NS_LITERAL_STRING("component_2"), comp2);
    
    nsCOMPtr<nsIDOMElement> importedComponent;
    imported.Get(comp1, getter_AddRefs(importedComponent));
    if (importedComponent == nsnull)
      continue;
    nsAutoString varName;
    importedComponent->GetAttribute(NS_LITERAL_STRING("name"), varName);
    mapel->SetAttribute(NS_LITERAL_STRING("component_1"), varName);

    imported.Get(comp2, getter_AddRefs(importedComponent));
    if (importedComponent == nsnull)
      continue;
    importedComponent->GetAttribute(NS_LITERAL_STRING("name"), varName);
    mapel->SetAttribute(NS_LITERAL_STRING("component_2"), varName);

    allConnections.AppendObject(conn);
  }
  return NS_OK;
}

nsresult
mozCellML::ImportLoaded(nsIDOMElement* importNode,
                        nsIDOMDocument* importedModel)
{
  nsresult rv;
  // We have just satisfied an import element. Look for components to import...
  nsCOMPtr<nsIDOMElement> importedRoot;
  importedModel->GetDocumentElement(getter_AddRefs(importedRoot));
  if (importedRoot == nsnull)
  {
    printf("Imported root element is null!\n");
    return NS_ERROR_FAILURE;
  }

  // Build a hashtable of all available components...
  nsInterfaceHashtable<nsStringHashKey, nsIDOMElement> AllComponents;
  nsInterfaceHashtable<nsStringHashKey,nsIDOMElement> ImportComponents;
  AllComponents.Init();
  ImportComponents.Init();
  HashAvailableComponents(AllComponents, importedRoot);

  nsCOMArray<nsIDOMElement> ComponentStack;

  // Seed the list of components we want with those requested directly, and
  // import the componets on the fly...
  rv = AddDirectImports(AllComponents, ImportComponents, importNode, ComponentStack);
  NS_ENSURE_SUCCESS(rv, rv);

  // Now, we expand the list to include encapsulated components.
  ExpandEncapsulatedList(ComponentStack, AllComponents, ImportComponents, importedRoot);

  // We also need to add the connections involving the imported components...
  AddRelevantConnections(importedRoot, ImportComponents);

  return NS_OK;
}

nsresult
mozCellML::InstallComponent(nsIDOMElement* componentEl,
                            nsAString& componentName)
{
  allComponents->AppendElement(componentEl, PR_FALSE);
  nsCOMPtr<nsIDOMNodeList> vl;
  componentEl->GetChildNodes(getter_AddRefs(vl));
  nsCOMPtr<nsIArray> variables;
  nsCOMPtr<nsISimpleEnumerator> variablesE;
  
  if (componentName.Equals(NS_LITERAL_STRING("")))
    componentEl->GetAttribute(NS_LITERAL_STRING("name"), componentName);
  else
    componentEl->SetAttribute(NS_LITERAL_STRING("name"), componentName);

  nsCAutoString ccomponentName;
  CopyUTF16toUTF8(componentName, ccomponentName);
#ifdef DEBUG_MODE
  printf("Installing component %s\n", ccomponentName.get());
#endif

  FilterNodeListByNameAndTypeNS2(
                                 vl,
                                 getter_AddRefs(variables),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("variable")
                                );

  size_t j, n;
  variables->GetLength(&n);
  variables->Enumerate(getter_AddRefs(variablesE));
  
  for (
       j = 0;
       j < n;
       j++
      )
  {
    // They will own us, because we must outlive them to protect doc...
    nsCOMPtr<nsISupports> variableSupports;
    variablesE->GetNext(getter_AddRefs(variableSupports));
    nsCOMPtr<nsIDOMElement> variableEl(do_QueryInterface(variableSupports));
    nsCOMPtr<mozCellMLVariable> v = new mozCellMLVariable(variableEl, this);
    PRBool isflagged;
    v->IsFlagged(mozICellML::NO_INTERFACE_IN, &isflagged);
    if (!isflagged)
      continue;
    variableList->AppendElement(v, PR_FALSE);
    nsAutoString name;
    v->GetName(name);
    allVariables.Put(componentName + NS_LITERAL_STRING("::") + name, v);
  }
  return NS_OK;
}

nsresult
mozCellML::FindVariableElements(nsIDOMElement* root)
{
  nsCOMPtr<nsIDOMNodeList> dnl;
  root->GetChildNodes(getter_AddRefs(dnl));
  nsCOMPtr<nsIArray> components;
  nsCOMPtr<nsISimpleEnumerator> componentsE;
  FilterNodeListByNameAndTypeNS2(
                                 dnl,
                                 getter_AddRefs(components),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("component")
                                );
  PRUint32 i, l;
  components->GetLength(&l);
  components->Enumerate(getter_AddRefs(componentsE));
  for (
       i = 0;
       i < l;
       i++
      )
  {
    nsCOMPtr<nsISupports> componentSupports;
    componentsE->GetNext(getter_AddRefs(componentSupports));
    nsCOMPtr<nsIDOMElement> componentEl(do_QueryInterface(componentSupports));

    nsAutoString compName;
    InstallComponent(componentEl, compName);
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::GetAllVariables(
                           PRUint32 flags,
                           nsISimpleEnumerator** enumerator
                          )
{
  if (doc == NULL)
    return NS_ERROR_NOT_INITIALIZED;

  nsCOMPtr<nsIMutableArray> allMatchingVars;
  NS_NewArray(getter_AddRefs(allMatchingVars));

  nsCOMPtr<nsISimpleEnumerator> allEnum;
  variableList->Enumerate(getter_AddRefs(allEnum));
  PRUint32 i, l;
  variableList->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    PRBool isflagged;
    nsCOMPtr<nsISupports> vSupports;
    allEnum->GetNext(getter_AddRefs(vSupports));
    nsCOMPtr<mozCellMLVariable> v(do_QueryInterface(vSupports));
    v->IsFlagged(flags, &isflagged);
    if (!isflagged)
      continue;
    allMatchingVars->AppendElement(v, PR_FALSE);
  }

  allMatchingVars->Enumerate(enumerator);
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::GetVariable(
                       const nsAString& name,
                       mozCellMLIVariable** var
                      )
{
  if (doc == NULL)
    return NS_ERROR_NOT_INITIALIZED;

  nsCOMPtr<nsIMutableArray> allMatchingVars;
  NS_NewArray(getter_AddRefs(allMatchingVars));

  nsCOMPtr<nsISimpleEnumerator> allEnum;
  variableList->Enumerate(getter_AddRefs(allEnum));
  PRUint32 i, l;
  variableList->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    nsAutoString vname;
    nsCOMPtr<nsISupports> vSupports;
    allEnum->GetNext(getter_AddRefs(vSupports));
    nsCOMPtr<mozCellMLVariable> v(do_QueryInterface(vSupports));
    v->GetName(vname);
    if (!name.Equals(vname))
      continue;
    *var = v;
    NS_ADDREF(*var);
    return NS_OK;
  }
  return NS_ERROR_FAILURE;
}

NS_IMETHODIMP
mozCellML::CheckValidity(
                         nsAString& validityError
                        )
{
  if (localValidationError)
  {
    validityError.Assign(validationErrorMessage);
    return NS_OK;
  }
  validityError.Assign(NS_LITERAL_STRING(""));
  return NS_OK;
}

nsresult
mozCellML::ValidationError(const nsAString& msg)
{
  localValidationError = PR_TRUE;
#if 0
  if (!validationErrorMessage.Equals(NS_LITERAL_STRING("")))
    validationErrorMessage.Append(NS_LITERAL_STRING(" and "));
  validationErrorMessage.Append(msg);
#else
  validationErrorMessage.Assign(msg);
#endif
#ifdef DEBUG_MODE
  nsCAutoString ve;
  CopyUTF16toUTF8(msg, ve);
  printf("Got a validation error: %s\n", ve.get());
#endif
  return NS_ERROR_INVALID_ARG;
}

nsresult
mozCellML::WriteDifferentialAssignment(
                                       PRFileDesc* prfd,
                                       PRUint32 i,
                                       mozCellMLVariable* v,
                                       nsInterfaceHashtable<nsStringHashKey,
                                       mozCellMLVariable>& allVariables
                                      )
{
  nsCAutoString code;
  nsresult ret;
  if ((ret = v->GetComputationString(this, allVariables, code)) != NS_OK)
    return ret;
  PR_fprintf(prfd, "  dx[%u] = %s;\n",
             i, code.get());
  return NS_OK;
}

nsresult
mozCellML::WriteSimpleEvaluation(
                                 PRFileDesc* prfd,
                                 PRUint32 i,
                                 mozCellMLVariable* v,
                                 nsInterfaceHashtable<nsStringHashKey,
                                 mozCellMLVariable>& allVariables
                                )
{
  nsCAutoString code;
  nsresult ret;
  if ((ret = v->GetComputationString(this, allVariables, code)) != NS_OK)
    return ret;
  PR_fprintf(prfd, "  x[%u] = %s;\n",
             i, code.get());
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::ParseCommandLine(const char* data, char** args, PRUint32 maxargs)
{
  PRUint32 n = 0;
  PRBool quoted = PR_FALSE, escaped = PR_FALSE;
  nsCAutoString tmp;

  maxargs -= 2;

  while (*data != 0)
  {
    if (escaped)
    {
      tmp += *data;
      escaped = PR_FALSE;
      data++;
      continue;
    }
#if 0
    else if (*data == '\\')
    {
      escaped = PR_TRUE;
      data++;
      continue;
    }
#endif
    else if (*data == '\"')
      quoted = !quoted;
    else if (!quoted && *data == ' ')
    {
      if (n < maxargs)
        args[n++] = ToNewCString(tmp);
      tmp = "";
    }
    else
      tmp += *data;
    data++;
  }

  args[n++] = ToNewCString(tmp);
  args[n++] = NULL;

  return NS_OK;
}

NS_IMETHODIMP
mozCellML::FreeCommandLine(char **args)
{
  while (*args)
  {
    PR_Free(*args);
    args++;
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::NewIntegrator(
                         const nsACString& cfile,
                         const nsACString& compile,
                         const nsACString& execute,
                         mozCellMLIIntegrator** result
                        )
{
#ifdef DEBUG_MODE
  printf("NewIntegrator called.\n");
#endif
  // At this stage, we actually generate code. The first step is to open a
  // C file to write the code to.

  nsCAutoString xcfile(cfile);
  PRFileDesc* prfd = PR_Open(xcfile.get(), PR_WRONLY | PR_CREATE_FILE,
                             0700);
#ifdef DEBUG_MODE
  printf("Creating ServerSocket...\n");
#endif

  PRUint32 i, l;

  if (prfd == NULL)
    return NS_ERROR_FAILURE;

  nsCOMPtr<nsIServerSocket> serverSocket = do_CreateInstance(NS_SERVERSOCKET_CONTRACTID);

  serverSocket->Init(0, PR_TRUE, 1);
  PRInt32 local_port;
  serverSocket->GetPort(&local_port);

#ifdef DEBUG_MODE
  printf("Writing integrator code...\n");
#endif

  // We have to find all the "hidden higher" variables...
  PRUint32 alloc_range = differentials.Count();
  l = differentials.Count();
  for (i = 0; i < l; i++)
  {
    PRUint32 vardeg = differentials[i]->FetchDegree();
    if (vardeg != 1)
    {
      differentials[i]->SetHigherSpace(alloc_range + order.Count(), alloc_range);
      alloc_range += (vardeg - 1);
    }
  }
  PRUint32 hidden_higher = alloc_range - differentials.Count();

  PR_fprintf(
             prfd,
             "#define ND %u\n"
             "#define NS %u\n\n"
             "#define CONNECT_PORT %u\n\n"
             // This also includes math.h or whatever is required.
             "#include <IntegrationHelpers.c>\n\n"
             "static inline void\n"
             "ComputeDifferentials(double y, const double *x, double *dx)\n"
             "{\n",
             differentials.Count() + hidden_higher,
             order.Count(),
             local_port
            );
  nsresult ret;
  for (i = 0, l = differentials.Count();
       i < l;
       i++)
    if ((ret = WriteDifferentialAssignment(prfd, i,
                                           differentials[i], allVariables))
        != NS_OK)
    {
#ifdef DEBUG_MODE
      printf("Error writing differential assignment.\n");
#endif
      PR_Close(prfd);
      return ret;
    }
  PR_fprintf(
             prfd,
             "}\n\n"
             "static inline void\n"
             "EvaluateSimple(double y, double *x)\n"
             "{\n"
            );
  for (
       i = 0, l = order.Count();
       i < l;
       i++
      )
    if ((ret = WriteSimpleEvaluation(prfd, i, order[i], allVariables)) !=
        NS_OK)
    {
#ifdef DEBUG_MODE
      printf("Error writing simple assignment.\n");
#endif
      PR_Close(prfd);
      return ret;
    }
  PR_fprintf(
             prfd,
             "}\n\n"
             "#include <DoIntegration.c>\n"
            );
#ifdef DEBUG_MODE
  printf("Closing C file...\n");
#endif
  PR_Close(prfd);

#ifdef DEBUG_MODE
  printf("Parsing compile command line...\n");
#endif
  char* argv[100];
  nsCAutoString xcompile(compile);
  ParseCommandLine(xcompile.get(), argv, 100);
  if (argv[0] == NULL)
    return NS_ERROR_FAILURE;
#ifdef DEBUG_MODE
  printf("Creating compiler process...\n");
#endif
  PRProcess* p = PR_CreateProcess(argv[0], argv, NULL, NULL);
  FreeCommandLine(argv);
  if (p == NULL)
    return NS_ERROR_FAILURE;
#ifdef DEBUG_MODE
  printf("Waiting for compilation to complete...\n");
#endif

  PRInt32 excode;
  PR_WaitProcess(p, &excode);
  if (excode != 0)
  {
#ifdef DEBUG_MODE
    printf("Compilation failed.\n");
#endif
    return NS_ERROR_FAILURE;
  }
#ifdef DEBUG_MODE
  printf("Compiled the program. Executing it...\n");
#endif

  nsCAutoString xexecute(execute);

  ParseCommandLine(xexecute.get(), argv, 100);
  if (argv[0] == NULL)
  {
#ifdef DEBUG_MODE
    printf("Bad execute command line.\n");
#endif
    return NS_ERROR_FAILURE;
  }

#ifndef MANUAL_RUN
  p = PR_CreateProcess(argv[0], argv, NULL, NULL);
#endif
  nsCAutoString exec(argv[0]);

  if (p == NULL)
  {
#ifdef DEBUG_MODE
    printf("Can't create process %s\n", argv[0]);
#endif
    FreeCommandLine(argv);
    return NS_ERROR_FAILURE;
  }
  FreeCommandLine(argv);

#ifdef DEBUG_MODE
  printf("Detaching from the program...\n");
#endif

#ifndef MANUAL_RUN
  PR_DetachProcess(p);
#endif

#ifdef DEBUG_MODE
  printf("Creating the integrator...\n");
#endif

  // We now have to create the integrator...
  *result = new mozCellMLIntegrator(
                                    serverSocket,
                                    differentials,
                                    order,
                                    exec,
                                    hidden_higher
                                   );
  NS_ADDREF(*result);
#ifdef DEBUG_MODE
  printf("Everything worked.\n");
#endif

  return NS_OK;
}
